use std::collections::HashMap;

use serde::{Deserialize, Serialize};
use serde_json::Value;
use serde_repr::Deserialize_repr;
use strum::{Display, EnumString};

// Use serde_repr to automatically handle integer representation
#[derive(Debug, PartialEq, Clone, Deserialize_repr, Serialize)]
#[repr(u8)]
#[allow(dead_code)]
pub enum TypPracoviste {
    Kraj = 1,
    Pracoviste = 2,
    CUZKSCD = 3,
    Inspektorat = 4,
    UK = 5,
    CUZK = 7,
}

#[derive(EnumString, Display)]
pub enum InspireLayer {
    #[strum(serialize = "ad")]
    Address,

    #[strum(serialize = "au")]
    AdministrativeUnit,

    #[strum(serialize = "cp")]
    CadastralParcel,

    #[strum(serialize = "cpx")]
    CadastralParcelExtended,

    #[strum(serialize = "bu")]
    Building,
}

/// Represents a row in the Katastralní Pracoviště CSV file
#[derive(Debug, Deserialize, Serialize)]
#[allow(dead_code)]
pub struct KatastralniPracovisteCsvRow {
    /// Kód pracoviště
    #[serde(rename = "KOD")]
    pub kod: u32,

    /// Zkratka
    #[serde(rename = "ZKRATKA")]
    pub zkratka: String,

    /// Typ pracoviště
    #[serde(rename = "TYP_PRAC")]
    pub typ_prac: TypPracoviste,

    /// Název pracoviště
    #[serde(rename = "NAZEV")]
    pub nazev: String,

    /// Zkrácený název pracoviště
    #[serde(rename = "NAZEV_ZKRACENY")]
    pub nazev_zkraceny: String,

    /// Nadřízené pracoviště
    #[serde(rename = "NADRIZ_PRAC")]
    pub nadriz_prac: String,

    /// Datum vzniku pracoviště
    #[serde(rename = "PLATNOST_OD")]
    pub platnost_od: String,

    /// Datum zániku pracoviště
    #[serde(rename = "PLATNOST_DO")]
    pub platnost_do: Option<String>,

    /// IČO
    #[serde(rename = "ICO")]
    pub ico: String,

    /// Telefonní číslo
    #[serde(rename = "TELEFON")]
    pub telefon: String,

    /// Číslo faxu
    #[serde(rename = "FAX")]
    pub fax: String,

    /// E-mailová adresa
    #[serde(rename = "EMAIL")]
    pub email: String,

    /// Kvalifikátor adresy pracoviště
    #[serde(rename = "TYP_ADRESY")]
    pub typ_adresy: String,

    /// Adresa – okres
    #[serde(rename = "OKRES")]
    pub okres: String,

    /// Adresa – obec
    #[serde(rename = "OBEC")]
    pub obec: String,

    /// Adresa – městský obvod Prahy
    #[serde(rename = "MESTSKY_OBVOD")]
    pub mestsky_obvod: Option<String>,

    /// Adresa – část obce
    #[serde(rename = "CAST_OBCE")]
    pub cast_obce: String,

    /// Adresa – číslo popisné nebo evidenční
    #[serde(rename = "CISLO_DOMOVNI")]
    pub cislo_domovni: String,

    /// Adresa – název ulice
    #[serde(rename = "NAZEV_ULICE")]
    pub nazev_ulice: String,

    /// Adresa – číslo orientační
    #[serde(rename = "CISLO_ORIENTACNI")]
    pub cislo_orientacni: String,

    /// Adresa – PSČ
    #[serde(rename = "PSC")]
    pub psc: String,

    /// Adresa - stát
    #[serde(rename = "STAT")]
    pub stat: String,

    /// Kód pracoviště pro variabilní symbol
    #[serde(rename = "KOD_PRO_VS")]
    pub kod_pro_vs: Option<u32>,
}

/// Represents a row in the Katastrální Území CSV file
#[derive(Debug, Deserialize, Serialize)]
#[allow(dead_code)]
pub struct KatastralniUzemiCsvRow {
    /// Kód kraje
    #[serde(rename = "KRAJ_KOD")]
    pub kraj_kod: u32,

    /// Název kraje
    #[serde(rename = "KRAJ_NAZEV")]
    pub kraj_nazev: String,

    /// Kód okresu
    #[serde(rename = "OKRES_KOD")]
    pub okres_kod: Option<u32>,

    /// Kód okresu NUTS4
    #[serde(rename = "NUTS4")]
    pub nuts4: Option<String>,

    /// Název okresu
    #[serde(rename = "OKRES_NAZEV")]
    pub okres_nazev: String,

    /// Kód obce
    #[serde(rename = "OBEC_KOD")]
    pub obec_kod: u32,

    /// Název obce
    #[serde(rename = "OBEC_NAZEV")]
    pub obec_nazev: String,

    /// Kód katastrálního území
    #[serde(rename = "KU_KOD")]
    pub ku_kod: u32,

    /// Pracovní číslo katastrálního území
    #[serde(rename = "KU_PRAC")]
    pub ku_prac: Option<u32>,

    /// Název katastrálního území
    #[serde(rename = "KU_NAZEV")]
    pub ku_nazev: String,

    /// Typ mapy
    #[serde(rename = "MAPA")]
    pub mapa: String,

    /// Číselná řada parcel
    #[serde(rename = "CISELNA_RADA")]
    pub ciselna_rada: u32,

    /// Datum vzniku katastrálního území
    #[serde(rename = "PLATNOST_OD")]
    pub platnost_od: String,

    /// Datum zániku katastrálního území
    #[serde(rename = "PLATNOST_DO")]
    pub platnost_do: Option<String>,

    /// Kód pracoviště
    #[serde(rename = "PRARES_KOD")]
    pub prares_kod: u32,

    /// Název pracoviště
    #[serde(rename = "PRARES_NAZEV")]
    pub prares_nazev: String,

    /// Poznámka
    #[serde(rename = "POZNAMKA")]
    pub poznamka: Option<String>,
}

impl From<&KatastralniUzemiCsvRow> for HashMap<String, String> {
    fn from(row: &KatastralniUzemiCsvRow) -> Self {
        let value = serde_json::to_value(row).unwrap();
        value
            .as_object()
            .unwrap()
            .iter()
            .map(|(k, v)| {
                (
                    k.to_lowercase(),
                    match v {
                        Value::Null => "".to_string(),
                        Value::String(s) => s.clone(),
                        _ => v.to_string(),
                    },
                )
            })
            .collect()
    }
}

#[derive(Debug, Deserialize)]
#[allow(dead_code)]
pub struct KuTableRow {
    pub pracoviste: String,
    pub kod_ku: String,
    pub pracoviste_cislo: String,
    pub nazev: String,
    pub dkm: u8,
    pub kmd: u8,
    pub kmd_dash: u8,
}
