use anyhow::Result;
use clap::{Parser, Subcommand, Args};
use csv::ReaderBuilder;
use gdal::{
    spatial_ref::{CoordTransform, SpatialRef},
    vector::LayerAccess,
    vsi,
};
use indicatif::{ProgressBar, ProgressState, ProgressStyle};
use rayon::prelude::*;
use reqwest::blocking::Client as ReqwestClient;
use serde::Deserialize;
use std::io::Cursor;
use std::time::Duration;
use zip::ZipArchive;

use encoding_rs::WINDOWS_1250;

mod cuzk_types;
mod tippecanoe;

fn fetch_csv_data<T>(url: &str) -> Result<Vec<T>>
where
    T: for<'de> Deserialize<'de>,
{
    let response = reqwest::blocking::get(url)?.bytes()?;
    let csv_text = unzip_csv(&response)?;
    let parsed_data = parse_csv::<T>(&csv_text)?;
    Ok(parsed_data)
}

fn unzip_csv(zip_bytes: &[u8]) -> Result<String> {
    let reader = Cursor::new(zip_bytes);
    let mut zip = ZipArchive::new(reader)?;
    let mut file = zip.by_index(0)?;
    let mut csv_data = Vec::new();
    std::io::copy(&mut file, &mut csv_data)?;
    let (decoded_str, _, _) = WINDOWS_1250.decode(&csv_data);
    Ok(decoded_str.into_owned())
}

fn parse_csv<T>(csv_text: &str) -> Result<Vec<T>>
where
    T: for<'de> Deserialize<'de>,
{
    let mut rdr = ReaderBuilder::new()
        .delimiter(b';')
        .has_headers(true)
        .from_reader(csv_text.as_bytes());

    let mut records = Vec::new();
    for result in rdr.deserialize() {
        let record: T = result?;
        records.push(record);
    }
    Ok(records)
}

#[allow(dead_code)]
fn ku_geo_url(dataset: cuzk_types::InspireLayer, ku_code: u32) -> String {
    format!(
        "https://services.cuzk.cz/gml/inspire/{}/epsg-4258/{}.zip",
        dataset, ku_code
    )
}

#[allow(dead_code)]
fn fetch_kn_prac() -> Result<Vec<cuzk_types::KatastralniPracovisteCsvRow>> {
    let url = "https://services.cuzk.cz/sestavy/cis/SC_PRACRES_DOTAZ.zip";
    fetch_csv_data::<cuzk_types::KatastralniPracovisteCsvRow>(url)
}

#[allow(dead_code)]
fn fetch_kn_uzemi() -> Result<Vec<cuzk_types::KatastralniUzemiCsvRow>> {
    let url = "https://services.cuzk.cz/sestavy/cis/SC_SEZNAMKUKRA_DOTAZ.zip";
    fetch_csv_data::<cuzk_types::KatastralniUzemiCsvRow>(url)
}

#[allow(dead_code)]
fn download_ku_geo_data(
    dataset: cuzk_types::InspireLayer,
    ku_code: u32,
    client: &ReqwestClient,
) -> Result<Vec<u8>> {
    let url = ku_geo_url(dataset, ku_code);
    let response = client.get(&url).send()?;
    let bytes = response.bytes()?;
    let reader = std::io::Cursor::new(bytes);
    let mut zip = zip::ZipArchive::new(reader)?;
    let mut geo_file = zip.by_index(0)?;
    let mut buffer = Vec::new();
    std::io::copy(&mut geo_file, &mut buffer)?;
    Ok(buffer)
}

fn process_gml_file(
    vsipath: &str,
    client: &mut postgres::Client,
    additional_properties: Option<std::collections::HashMap<String, String>>,
) -> Result<()> {
    let dataset = gdal::Dataset::open(vsipath)?;
    let mut cadastral_parcel = dataset.layer_by_name("CadastralParcel")?;
    let htransform =
        CoordTransform::new(&SpatialRef::from_epsg(4258)?, &SpatialRef::from_epsg(4326)?)?;

    let mut wkb_values: Vec<Vec<u8>> = Vec::new();
    let mut json_values: Vec<serde_json::Value> = Vec::new();

    for feature in cadastral_parcel.features() {
        let mut properties = std::collections::HashMap::new();
        feature.fields().for_each(|(name, value)| {
            let value_string = match value {
                Some(v) => v.into_string().unwrap_or_default(),
                None => String::new(),
            };
            properties.insert(name.to_string(), value_string);
        });

        if let Some(additional_properties) = &additional_properties {
            properties.extend(additional_properties.clone());
        }

        if let Some(geom) = feature.geometry() {
            if !geom.is_valid() {
                eprintln!("❌ Invalid input geometry");
            }
            let transformed_geom = geom.transform(&htransform)?;
            if !transformed_geom.is_valid() {
                eprintln!("❌ Invalid transformed geometry");
            }
            let wkb = transformed_geom.wkb()?;
            let json_val = serde_json::to_value(properties)?;
            wkb_values.push(wkb);
            json_values.push(json_val);
        }
    }

    let num_rows = wkb_values.len();
    if num_rows == 0 {
        dataset.close()?;
        return Ok(());
    }

    let mut values_sql = Vec::new();
    for i in 0..num_rows {
        let p1 = i * 2 + 1;
        let p2 = i * 2 + 2;
        values_sql.push(format!("(ST_GeomFromWKB(${},4326), ${})", p1, p2));
    }
    let query = format!(
        "INSERT INTO parcel (geom, properties) VALUES {}",
        values_sql.join(", ")
    );

    let mut params: Vec<&(dyn postgres::types::ToSql + Sync)> = Vec::with_capacity(num_rows * 2);
    for i in 0..num_rows {
        params.push(&wkb_values[i]);
        params.push(&json_values[i]);
    }

    let mut transaction = client.transaction()?;
    transaction.execute(&query, &params[..])?;
    transaction.commit()?;
    dataset.close()?;
    Ok(())
}

fn db_init(client: &mut postgres::Client) -> Result<()> {
    let _ = client.execute("CREATE EXTENSION IF NOT EXISTS postgis", &[])?;
    let _ = client.execute("CREATE EXTENSION IF NOT EXISTS pg_uuidv7", &[])?;
    Ok(())
}

fn db_create_client() -> Result<postgres::Client> {
    Ok(postgres::Client::connect(
        "host=127.0.0.1 user=postgres port=49264",
        postgres::NoTls,
    )?)
}

fn db_create_tables(client: &mut postgres::Client, drop: bool) -> Result<()> {
    if drop {
        client.execute("DROP TABLE IF EXISTS building", &[])?;
        client.execute("DROP TABLE IF EXISTS parcel", &[])?;
        client.execute("DROP TABLE IF EXISTS ku", &[])?;
        client.execute("DROP TABLE IF EXISTS ku_prac", &[])?;
    }
    client.execute(
        "CREATE TABLE IF NOT EXISTS ku_prac (
            id UUID PRIMARY KEY DEFAULT uuid_generate_v7(),
            code INTEGER NOT NULL,
            name TEXT NOT NULL,
            okres TEXT NOT NULL,
            obec TEXT NOT NULL,
            cast_obce TEXT NOT NULL,
            slug TEXT NOT NULL
        )",
        &[],
    )?;
    client.execute(
        "CREATE TABLE IF NOT EXISTS ku (
            id UUID PRIMARY KEY DEFAULT uuid_generate_v7(),
            code INTEGER NOT NULL,
            name TEXT NOT NULL,
            ku_prac_id UUID NOT NULL REFERENCES ku_prac(id)
        )",
        &[],
    )?;
    client.execute(
        "CREATE TABLE IF NOT EXISTS parcel (
            id UUID PRIMARY KEY DEFAULT uuid_generate_v7(),
            geom GEOMETRY(MultiPolygon, 4326) NOT NULL,
            properties JSONB NOT NULL DEFAULT '{}'
        )",
        &[],
    )?;
    client.execute(
        "CREATE TABLE IF NOT EXISTS building (
            id UUID PRIMARY KEY DEFAULT uuid_generate_v7(),
            geom GEOMETRY(MultiPolygon, 4326) NOT NULL,
            properties JSONB NOT NULL DEFAULT '{}'
        )",
        &[],
    )?;
    Ok(())
}

fn setup_env() {
    std::env::set_var("CPL_DEBUG", "OFF");
    std::env::set_var("CPL_LOG_ERRORS", "OFF");
    std::env::set_var("OGR_GEOMETRY_ACCEPT_UNCLOSED_RING", "NO");
    std::env::set_var("GML_SKIP_CORRUPTED_FEATURES", "YES");
    std::env::set_var("NAS_SKIP_CORRUPTED_FEATURES", "YES");
}

fn create_progress_bar(len: u64) -> ProgressBar {
    let pb = ProgressBar::new(len);
    let style = ProgressStyle::with_template(
        "[{elapsed_precise}] {bar:40.cyan/blue} {pos}/{len} ({eta}) (avg: {avg}) {msg}",
    )
    .unwrap()
    .with_key(
        "avg",
        |state: &ProgressState, w: &mut dyn std::fmt::Write| {
            let position = state.pos();
            if position > 0 {
                let avg_secs = state.elapsed().as_secs_f64() / position as f64;
                let avg_duration = Duration::from_secs_f64(avg_secs);
                write!(w, "{:.2?}", avg_duration).unwrap();
            } else {
                write!(w, "N/A").unwrap();
            }
        },
    );
    pb.set_style(style);
    pb
}

#[derive(Parser, Debug)]
#[command(name = "app")]
#[command(about = "CLI for upserting and listing entities", long_about = None)]
struct Cli {
    #[command(subcommand)]
    command: Commands,
}

#[derive(Subcommand, Debug)]
enum Commands {
    Upsert(UpsertCommand),
    List(ListCommand),
    Tiles,
}

#[derive(Args, Debug)]
struct UpsertCommand {
    #[command(subcommand)]
    target: UpsertTarget,
}

#[derive(Subcommand, Debug)]
enum UpsertTarget {
    Parcel {
        #[arg(long = "filter-kraj")]
        filter_kraj: Option<u32>,
        #[arg(long = "filter-okres")]
        filter_okres: Option<u32>,
    },
    Building
}

#[derive(Args, Debug)]
struct ListCommand {
    #[command(subcommand)]
    target: ListTarget,
}

#[derive(Subcommand, Debug)]
enum ListTarget {
    Pracoviste,
    KnAreas,
}

fn db_upsert(
    filter_kraj: Option<u32>,
    filter_okres: Option<u32>,
    // TODO: Be able to specify the inspire layer
    // layer: Option<cuzk_types::InspireLayer>,
) -> Result<()> {
    setup_env();
    let mut db = db_create_client()?;
    db_init(&mut db)?;
    db_create_tables(&mut db, true)?;
    db.execute("ALTER TABLE parcel SET UNLOGGED", &[])?;

    let uzemi = fetch_kn_uzemi()?;
    let ku_uzemis = uzemi
        .into_iter()
        .filter(|ku| {
            match (filter_kraj, filter_okres) {
                (Some(kraj), Some(okres)) => ku.kraj_kod == kraj && ku.okres_kod == Some(okres),
                (Some(kraj), None) => ku.kraj_kod == kraj,
                (None, Some(okres)) => ku.okres_kod == Some(okres),
                (None, None) => true,
            }
        })
        .collect::<Vec<_>>();

    let pb = create_progress_bar(ku_uzemis.len() as u64);

    let results: Result<Vec<()>> = ku_uzemis
        .par_iter()
        .map(|ku_uzemi| {
            let mut db = db_create_client()?;
            let http_client = ReqwestClient::builder().build()?;

            match download_ku_geo_data(
                cuzk_types::InspireLayer::CadastralParcel,
                ku_uzemi.ku_kod,
                &http_client,
            ) {
                Ok(buffer) => {
                    let vsi_path = format!("/vsimem/ku_{}.gml", ku_uzemi.ku_kod);
                    vsi::create_mem_file(&vsi_path, buffer)?;
                    match process_gml_file(vsi_path.as_str(), &mut db, Some(ku_uzemi.into())) {
                        Ok(_) => pb.println(format!("✅ {:?}", ku_uzemi.ku_nazev)),
                        Err(e) => {
                            pb.println(format!("❌ {:?}", ku_uzemi.ku_nazev));
                            eprintln!("Error: {:?}", e);
                        }
                    }
                    vsi::unlink_mem_file(vsi_path)?;
                }
                Err(e) => {
                    pb.println(format!("❌ {:?}", ku_uzemi.ku_nazev));
                    eprintln!("Error: {:?}", e);
                }
            }
            pb.inc(1);
            Ok(())
        })
        .collect();
    results?;

    db.execute("ALTER TABLE parcel SET LOGGED", &[])?;
    pb.finish_with_message("Processing complete");

    Ok(())
}

fn gen_tiles(okres_kod: u32) -> Result<()> {
    let mut db = db_create_client()?;
    let row = db.query_one("
        SELECT ST_AsFlatGeobuf(geo, false, 'geom') AS fgb
        FROM (
            SELECT geom
            FROM parcel
            WHERE properties->>'okres_kod' = $1
        ) AS geo;
        ",
        &[&okres_kod.to_string()],
    )?;

    let fgb_data: Vec<u8> = row.get("fgb");

    let fgb_path = format!("{}.fgb", okres_kod);
    let pmt_path = std::path::PathBuf::from(format!("{}.pmtiles", okres_kod));
    std::fs::write(&fgb_path, &fgb_data)?;
    tippecanoe::run_tippecanoe(&tippecanoe::TippecanoeConfig {
        input_file: std::path::Path::new(&format!("{}.fgb", okres_kod)),
        output_file: pmt_path,
        ..Default::default()
    })?;
    Ok(())
}

// Fix and report to cuzk: Polygon must have at least four points in each ring
fn main() -> Result<()> {
    let cli = Cli::parse();

    match cli.command {
        Commands::Upsert(upsert) => match upsert.target {
            UpsertTarget::Parcel {
                filter_kraj,
                filter_okres,
            } => {
                db_upsert(filter_kraj, filter_okres)?;
            }
            UpsertTarget::Building => todo!()
        },
        Commands::List(list) => match list.target {
            ListTarget::Pracoviste => {
                let pracoviste = fetch_kn_prac()?;
                let json = serde_json::to_string(&pracoviste)?;
                println!("{}", json);
            }
            ListTarget::KnAreas => {
                let uzemi = fetch_kn_uzemi()?;
                let json = serde_json::to_string(&uzemi)?;
                println!("{}", json);
            }
        },
        Commands::Tiles => {
            // TODO: Filtering
            gen_tiles(3807)?;
        }
    }

    Ok(())
}
