use std::process::{Command, Stdio};
use std::io::{self, BufRead};
use std::path::{Path, PathBuf};

#[derive(Debug)]
pub struct TippecanoeConfig<'a> {
    pub minimum_zoom: u8,
    pub maximum_zoom: u8,
    pub full_detail: u8,
    pub no_line_simplification: bool,
    pub no_tiny_polygon_reduction: bool,
    pub coalesce_smallest_as_needed: bool,
    pub simplify_only_low_zooms: bool,
    pub maximum_tile_bytes: u32,
    pub layer_name: &'a str,
    pub output_file: PathBuf,
    pub force: bool,
    pub input_file: &'a Path,
}

impl<'a> Default for TippecanoeConfig<'a> {
    fn default() -> Self {
        TippecanoeConfig {
            minimum_zoom: 11,
            maximum_zoom: 16,
            full_detail: 16,
            no_line_simplification: true,
            no_tiny_polygon_reduction: true,
            coalesce_smallest_as_needed: true,
            simplify_only_low_zooms: true,
            maximum_tile_bytes: 1_000_000,
            layer_name: "CadastralParcel",
            output_file: PathBuf::from("output.pmtiles"),
            force: true,
            input_file: Path::new("input.fgb"),
        }
    }
}

pub fn run_tippecanoe(config: &TippecanoeConfig) -> io::Result<()> {
    let mut args = vec![
        "--minimum-zoom".into(), config.minimum_zoom.to_string(),
        "--maximum-zoom".into(), config.maximum_zoom.to_string(),
        "--full-detail".into(), config.full_detail.to_string(),
        "--maximum-tile-bytes".into(), config.maximum_tile_bytes.to_string(),
        "-l".into(), config.layer_name.into(),
        "-o".into(), config.output_file.to_string_lossy().to_string(),
    ];

    if config.no_line_simplification {
        args.push("--no-line-simplification".into());
    }
    if config.no_tiny_polygon_reduction {
        args.push("--no-tiny-polygon-reduction".into());
    }
    if config.coalesce_smallest_as_needed {
        args.push("--coalesce-smallest-as-needed".into());
    }
    if config.simplify_only_low_zooms {
        args.push("--simplify-only-low-zooms".into());
    }
    if config.force {
        args.push("--force".into());
    }

    args.push(config.input_file.to_string_lossy().to_string());

    let mut child = Command::new("tippecanoe")
        .args(&args)
        .stdout(Stdio::piped())
        .stderr(Stdio::piped())
        .spawn()?;

    let stdout = child.stdout.take().unwrap();
    let stderr = child.stderr.take().unwrap();

    let stdout_thread = std::thread::spawn(move || {
        let reader = io::BufReader::new(stdout);
        for line in reader.lines() {
            if let Ok(l) = line {
                println!("[tippecanoe stdout] {}", l);
            }
        }
    });

    let stderr_thread = std::thread::spawn(move || {
        let reader = io::BufReader::new(stderr);
        for line in reader.lines() {
            if let Ok(l) = line {
                eprintln!("[tippecanoe stderr] {}", l);
            }
        }
    });

    let status = child.wait()?;
    stdout_thread.join().unwrap();
    stderr_thread.join().unwrap();

    if status.success() {
        Ok(())
    } else {
        Err(io::Error::new(io::ErrorKind::Other, "tippecanoe failed"))
    }
}
