#!/usr/bin/env nix-shell
#! nix-shell -i bash
#! nix-shell -p bash curl jq fd parallel gdal tippecanoe htmlq ripunzip

set -e

rm -rf ./data

mkdir -p data

KRAJ_CODE=${1:-100}

YEAR=$(date +%Y)

KU_CODES=$(
  curl -s "https://cuzk.gov.cz/Dokument.aspx?AKCE=META:SESTAVA:MDR001_XSLT:WEBCUZK_KRAJEKOD:${KRAJ_CODE}_PZEROK:${YEAR}" |
  htmlq -t 'tr.radek td:nth-child(2)'
)

URLS=$(
  curl -s 'https://services.cuzk.cz/gml/inspire/cp/epsg-4258/' |
  grep -oP '(?<=href=")[^"]+\.zip' |
  grep -f <(echo "$KU_CODES")
)

echo "Downloading files..."
echo "$URLS" | parallel -j 10 "ripunzip unzip-uri --output-directory data {}"

echo "Creating unified VRT file..."
cat > data/cadastral.vrt << EOF
<OGRVRTDataSource>
    <OGRVRTUnionLayer name="CadastralParcel">
EOF

echo "Adding individual XML files to VRT..."
find data -name "*.xml" | while read xml_file; do
  cat >> data/cadastral.vrt << EOF
        <OGRVRTLayer name="$(basename "$xml_file" .xml)">
            <SrcDataSource>$xml_file</SrcDataSource>
            <SrcLayer>CadastralParcel</SrcLayer>
        </OGRVRTLayer>
EOF
done

echo "Closing VRT..."
cat >> data/cadastral.vrt << EOF
    </OGRVRTUnionLayer>
</OGRVRTDataSource>
EOF

echo "Converting VRT to FlatGeobuf..."

FLAT_GEOBUF_FILE="data/${KRAJ_CODE}.fgb"

ogr2ogr -s_srs "EPSG:4258" -t_srs "EPSG:4326" -f FlatGeobuf "${FLAT_GEOBUF_FILE}" data/cadastral.vrt -skipfailures

echo "Removing temp files..."
fd -tf -e xml -e vrt -e gfs --full-path data -X rm

echo "File information:"
ogrinfo -so "${FLAT_GEOBUF_FILE}"

echo "Layer information:"
ogrinfo -so "${FLAT_GEOBUF_FILE}" CadastralParcel

echo "Sample data (first 5 features):"
ogrinfo "${FLAT_GEOBUF_FILE}" CadastralParcel -limit 1

echo "Creating a tileset with Tippecanoe..."
tippecanoe \
  --minimum-zoom 11 \
  --maximum-zoom 16 \
  --full-detail 16 \
  --no-line-simplification \
  --no-tiny-polygon-reduction \
  --coalesce-smallest-as-needed \
  --simplify-only-low-zooms \
  --maximum-tile-bytes 1000000 \
 -l CadastralParcel \
 -o "${KRAJ_CODE}.pmtiles" \
 --force "${FLAT_GEOBUF_FILE}"

echo "Removing temp files..."
rm -rf data

echo "Done"
exit 0

# 100 300 700 460 660 500 800 860 600 400 200 560 760 770
